import os
import re
import zipfile
import bpy

from distutils.version import LooseVersion
from enum import Enum, auto
from bpy_extras.io_utils import ImportHelper

from . import icons
from . import constants

class Errors(Enum):
    MC_NOT_FOUND = auto()
    NO_VERSION_INSTALLED = auto()
    MIN_VERSION_EXCEEDED = auto()
    VERSION_FILE_NOT_FOUND = auto()
    NOT_ALL_TEXTURES = auto()


    def error_text(self):
        descriptions = {
            Errors.MC_NOT_FOUND: "Minecraft could not be found",
            Errors.NO_VERSION_INSTALLED: "No Minecraft version installed",
            Errors.MIN_VERSION_EXCEEDED: f'Minimum version requirement [{MC_TEXTURES_LOAD_OT_SET.MIN_VERSION}] not met',
            Errors.VERSION_FILE_NOT_FOUND: "MC Version.jar file not found",
            Errors.NOT_ALL_TEXTURES: "Not all needed textures found"
            
        }
        return descriptions.get(self, "Unknown error.")


class MC_TEXTURES_LOAD_OT_SET(bpy.types.Operator):
    bl_idname = "thomasriglegacy.mc_textures_import"
    bl_label = "load mc textures"
    bl_description = ""

    MIN_VERSION = [1, 21]
    error = None
    jar_path = None

    def execute(self, context):
        preferences = context.preferences.addons[constants.PACKAGE].preferences

        status = self.get_jar_path()
        if status:
            try:
                self.extract_from_zip()
                if self.error:
                    preferences.mc_textures_ignore = True
                    self.report({"WARNING"}, self.error.error_text())
                    bpy.ops.wm.save_userpref()
                    return{'CANCELLED'}
            except FileNotFoundError:
                preferences.mc_textures_ignore = True
                self.error = Errors.VERSION_FILE_NOT_FOUND
                self.report({"WARNING"}, self.error.error_text())
                bpy.ops.wm.save_userpref()
                return{'CANCELLED'}
        else:
            preferences.mc_textures_ignore = True
            self.report({"WARNING"}, self.error.error_text())
            bpy.ops.wm.save_userpref()
            return{'CANCELLED'}
        
        # set addon preferences texture loaded property
        preferences.mc_textures_loaded = True
        bpy.ops.wm.save_userpref()

        # reload icons
        icons.unregister()
        icons.register()
        version = os.path.splitext(os.path.basename(self.jar_path))[0]
        self.report({"INFO"}, version + " MC textures loaded successfully")
        return{'FINISHED'}
    
    def get_jar_path(self) -> bool:
        # Minecraft jar files directory
        if os.name == 'nt':  # Windows
            minecraft_dir = os.path.join(os.getenv('APPDATA'), '.minecraft', 'versions')
        elif os.name == 'posix':  # macOS and Linux
            if 'darwin' in os.uname().sysname.lower():  # macOS
                minecraft_dir = os.path.join(os.path.expanduser('~'), 'Library', 'Application Support', 'minecraft', 'versions')
            else:  # Linux
                minecraft_dir = os.path.join(os.path.expanduser('~'), '.minecraft', 'versions')

        if not os.path.exists(minecraft_dir): # MC not found
            self.error = Errors.MC_NOT_FOUND
            return False
        
        # get latest "normal" minecraft versions
        versions = os.listdir(minecraft_dir)
        if len(versions) == 0:
            self.error = Errors.NO_VERSION_INSTALLED
            return False
        
        # filter out snapshots and mods
        versions = [version for version in versions if not re.search('[a-zA-Z]', version)]
        # sort versions from low to high
        versions = sorted(versions, key=LooseVersion)
        for version in reversed(versions):
            # check min version
            int_version = [int(v) for v in version.split('.')]
            not_sufficient = int_version[0] < self.MIN_VERSION[0] or int_version[1] < self.MIN_VERSION[1]

            if not_sufficient:
                self.error = Errors.MIN_VERSION_EXCEEDED
                return False

            # check if version.jar exists
            file_name = version + '.jar'
            jar_path = os.path.join(minecraft_dir, version, file_name)
            if os.path.exists(jar_path):
                self.jar_path = jar_path
                return True
        return False

    def extract_from_zip(self) -> None:
        texture_path = bpy.utils.extension_path_user(package = constants.PACKAGE, path = "textures", create=True)
        textures, extracted_textures = extract_from_zip(texture_path, self.jar_path)

        if not textures == extracted_textures:
            self.error = Errors.NOT_ALL_TEXTURES
            return False


class MC_TEXTURES_IMPORT_OT_SET(bpy.types.Operator, ImportHelper):
    bl_idname = "thomasriglegacy.mc_textures_import_manually" 
    bl_label = "import mc textures"

    filter_glob : bpy.props.StringProperty(
        default = "*.zip;*.rar;*.jar",
        options = {"HIDDEN"}
        ) # type: ignore

    def execute(self, context):
        filepath = self.filepath
        preferences = context.preferences.addons[constants.PACKAGE].preferences

        texture_path = bpy.utils.extension_path_user(package = constants.PACKAGE, path = "textures", create=True)
        textures, extracted_textures = extract_from_zip(texture_path, filepath)

        if not textures == extracted_textures:
            self.error = Errors.NOT_ALL_TEXTURES
            preferences.mc_textures_ignore = True
            self.report({"WARNING"}, self.error.error_text())
            bpy.ops.wm.save_userpref()
            return{'CANCELLED'}
        
        # set addon preferences texture loaded property
        preferences.mc_textures_loaded = True
        bpy.ops.wm.save_userpref()

        # reload icons
        icons.unregister()
        icons.register()
        version = os.path.splitext(os.path.basename(filepath))[0]
        self.report({"INFO"}, version + " MC textures loaded successfully")
        return{'FINISHED'}


class MC_TEXTURES_SKIP_OT_SET(bpy.types.Operator):
    bl_idname = "thomasriglegacy.mc_textures_skip"
    bl_label = "skip"
    bl_description = ""

    def execute(self, context):
        preferences = context.preferences.addons[constants.PACKAGE].preferences
        preferences.mc_textures_ignore = True
        bpy.ops.wm.save_userpref()
        return{'FINISHED'}


def extract_from_zip(texture_path, jar_path):
    trims_zip_dir = os.path.join(texture_path, "trims")
    armor_dir = os.path.join(texture_path, "armor")
    icon_dir = os.path.join(texture_path, "icons")

    textures = {
        icon_dir : ["textures/item/iron_chestplate.png", "textures/item/elytra.png", "textures/item/glow_item_frame.png"],
        texture_path : ["entity/elytra.png"],
        trims_zip_dir : ["textures/trims/models/armor"],
        armor_dir : ["textures/models/armor"]
    }

    extracted_textures = {}

    with zipfile.ZipFile(jar_path) as archive:
        for file in archive.namelist():
            for dir, value in textures.items():
                if any(texture in file for texture in value):
                    target_name = os.path.basename(file)
                    target_path = os.path.join(dir, target_name)  # output path

                    # create target dummy file
                    os.makedirs(os.path.dirname(target_path), exist_ok = True)
                    with open(target_path, "wb") as f:  # open the output path for writing
                        f.write(archive.read(file)) # write file to target

                    # extracted textures
                    extracted_textures[dir] = value
                    break
    return textures, extracted_textures

def register():
    bpy.utils.register_class(MC_TEXTURES_LOAD_OT_SET)
    bpy.utils.register_class(MC_TEXTURES_IMPORT_OT_SET)
    bpy.utils.register_class(MC_TEXTURES_SKIP_OT_SET)

def unregister():
    bpy.utils.unregister_class(MC_TEXTURES_SKIP_OT_SET)
    bpy.utils.unregister_class(MC_TEXTURES_IMPORT_OT_SET)
    bpy.utils.unregister_class(MC_TEXTURES_LOAD_OT_SET)