import bpy
from bpy.props import BoolProperty, EnumProperty,FloatVectorProperty
import bmesh
import os

from .. import utils
from .. import icons
from .. import constants

class THOMAS_RIG_ARMOR_ADD(bpy.types.Operator):
    bl_idname = "thomasriglegacy.addarmor"
    bl_label = "add armor"
    bl_options = {'REGISTER', 'UNDO'}

    def create_material_enum_property(helmet = False):
        keys = ["leather", "chainmail", "iron", "gold", "diamond", "netherite"]
        if helmet:
            keys.append("turtle")

        return EnumProperty(
            name = "",
            default = 'diamond',
            items = [utils.create_item(key) for key in keys]
        )

    def create_trim_enum_property():
        keys = ("None", "bolt", "coast", "dune", "eye", "flow", "host", 
                "raiser", "rib", "sentry","shaper", "silence", "snout",
                "spire", "tide", "vex", "ward", "wayfinder", "wild")
        
        return EnumProperty(
            name = '',
            default = 'None',
            items = [utils.create_item(key) for key in keys]
        )

    def create_trim_colour_property():
        return FloatVectorProperty(
            name="", 
            subtype='COLOR', 
            size = 4,
            default=[1, 1, 1, 1],
            max = 1,
            min = 0,
        )

    # loaded mc textures
    loaded = False
    
    # armor parts
    helmet : BoolProperty() # type: ignore
    chestplate : BoolProperty() # type: ignore
    leggings : BoolProperty() # type: ignore
    boots : BoolProperty() # type: ignore

    armor_parts = ["helmet", "chestplate", "leggings", "boots"]

    # armor select
    armor_type : EnumProperty(
        default = 'default',
        items =  [utils.create_item(key) for key in ("default", "custom")]
    ) # type: ignore

    # custom armor
    armor_custom_type : EnumProperty(
        default = 'samurai',
        items =  [utils.create_item(key) for key in ("samurai", "cultist", "zombiePlate")]
    ) # type: ignore

    # default armor
    helmet_material : create_material_enum_property(True) # type: ignore
    chestplate_material : create_material_enum_property() # type: ignore
    leggings_material : create_material_enum_property() # type: ignore
    boots_material : create_material_enum_property() # type: ignore

    # armor trims
    helmet_trim : create_trim_enum_property() # type: ignore
    chestplate_trim : create_trim_enum_property() # type: ignore
    leggings_trim : create_trim_enum_property() # type: ignore
    boots_trim : create_trim_enum_property() # type: ignore

    # armor trims colour
    helmet_trim_colour : create_trim_colour_property() # type: ignore
    chestplate_trim_colour : create_trim_colour_property() # type: ignore
    leggings_trim_colour : create_trim_colour_property() # type: ignore
    boots_trim_colour : create_trim_colour_property() # type: ignore

    # alternative textures
    helmet_alt_texture : BoolProperty() # type: ignore
    chestplate_alt_texture : BoolProperty() # type: ignore
    leggings_alt_texture : BoolProperty() # type: ignore
    boots_alt_texture : BoolProperty() # type: ignore


    def invoke(self, context, event):
        # set mc textures loaded property and default for alternative textures
        preferences = context.preferences.addons[constants.PACKAGE].preferences
        self.loaded = preferences.mc_textures_loaded

        self.helmet_alt_texture = not self.loaded
        self.chestplate_alt_texture = not self.loaded
        self.leggings_alt_texture = not self.loaded
        self.boots_alt_texture = not self.loaded

        return context.window_manager.invoke_props_dialog(self, width=400)

    def draw(self, context):
        column = self.layout.column()
        column.row().prop(self, "armor_type", expand = True)

        pcoll = icons.thomas_icons["thomas_legacy"]
        if self.armor_type == "default":
            utils.UI_Utils.spacer(column, 0.3)

            icon = {
                'helmet' : pcoll["helmet_iron"].icon_id,
                'chestplate' : pcoll["chestplate_iron"].icon_id,
                'leggings' : pcoll["leggings_iron"].icon_id,
                'boots' : pcoll["boots_iron"].icon_id
            }
            
            for element in self.armor_parts:
                row = column.row()
                row.label(text="", icon_value=icon.get(element))
                row.prop(self, element)

                if getattr(self, element):
                    # alternative tetures
                    split = row.split()
                    split.enabled = self.loaded
                    split.prop(self, element+'_alt_texture', icon = "UV_SYNC_SELECT", text = "") # alt texture

                    # material
                    row.prop(self, element + "_material")

                    # trims: disabled if original mc textures not loaded
                    row = row.split()
                    row.enabled = self.loaded

                    row = row.row()
                    row.prop(self, element + "_trim")
                    row = row.row()
                    row.enabled = (getattr(self, f"{element}_trim") != 'None')
                    row.prop(self, element + "_trim_colour")
        else:
            column.template_icon_view(context.window_manager.thomas_rig_legacy, "custom_armor")
            
            utils.UI_Utils.spacer(column, 0.3)
            column.prop(self, "helmet")
            column.prop(self, "chestplate")
            column.prop(self, "leggings")
            
            if context.window_manager.thomas_rig_legacy.custom_armor[:-4] in {'Scuba'}:
                column.prop(self, "boots")

    def execute(self, context):
        if not any([self.helmet, self.chestplate, self.leggings, self.boots]):
            self.report({'WARNING'}, 'No armor part selected')
            return {'FINISHED'}
        try:
            mode = context.object.mode
        except:
            mode = False

        self.rig = utils.get_rig()
        context.view_layer.objects.active = self.rig
        self.rig.data.pose_position = 'REST'

        if self.armor_type == "default":
            for piece in self.armor_parts:
                if not getattr(self, piece):
                    continue

                # default armor
                context.view_layer.objects.active = self.rig
                item = self.add_default_armor(piece)

                # default trim
                part = f'{piece}_trim'
                if  getattr(self, part) != 'None':
                    self.add_trims(piece, item)
                    ThomasRigBakeArmor.adjust_material(item)
                else:
                    ThomasRigBakeArmor.delete_nodes_no_trim(item)
                if getattr(self, f"{piece}_material") == 'leather':
                    ThomasRigBakeArmor.add_nodes_leather(item)
                ThomasRigBakeArmor.remove_alpha_faces(item)
        else:
            for piece in self.armor_parts:
                if getattr(self, piece):
                    self.add_custom_armor(piece)

        bpy.ops.object.select_all (action='DESELECT')
        context.view_layer.objects.active = self.rig
        if mode:
            bpy.ops.object.mode_set(mode=mode, toggle=False)
        
        # force update no face mode
        self.rig.data.update_tag()
        # set rest position
        self.rig.data.pose_position = 'POSE'
        return {'FINISHED'}
    
    def add_default_armor(self, part):
        default_armor_file = os.path.join(constants.ARMOR_PATH, "armor_default.blend")

        # append
        directory = os.path.join(default_armor_file, "Object")
        filepath = os.path.join(directory, part)
        bpy.ops.wm.append(
            filepath=filepath,
            filename=part,
            directory=directory,
            link=False,
            active_collection=True
        )
        
        bpy.ops.object.mode_set(mode = "OBJECT")
        item = bpy.context.selected_objects[0]
        item.location.x = 0
        item.location.y = 0
        # item.location.z = 0

        # set correct texture
        # get texture path
        if not getattr(self, part+"_alt_texture"):
            res_path = bpy.utils.extension_path_user(package = constants.PACKAGE, path = "textures")
            texture_types = os.path.join(res_path, 'armor')
        else:
            texture_types = constants.ARMOR_PATH_VANILLA

        # use second texture layer if leggings
        layer = '_layer_2' if part == "leggings" else '_layer_1'

        material = item.material_slots[0].material
        nodes = material.node_tree.nodes
        texture_type = getattr(self, f"{part}_material")
        texture = os.path.join(texture_types, texture_type + layer +'.png')
        image = bpy.data.images.load(texture)
        nodes['base'].image = image

        # modifiers
        # add armature modifier
        utils.add_modifier_armature(self.rig, item)

        # taper lattice
        self.add_taper(item, part)

        # add lattice SmartHead and solidify
        if "helmet" in part:
            utils.add_modifier_lattice_smart_deform(self.rig, item)
            utils.add_modifier_lattice_head(self.rig, item)
            utils.add_modifier_subdivision(item, 1, 3)
            utils.add_modifier_solidify(item, -0.075)

            bpy.context.view_layer.objects.active = item
        else:
            utils.add_modifier_subdivision(item, 1, 3)
            self.add_drivers_default_armor(item, self.rig, part)
        # add subdivisions
        return item

    def add_trims(self, piece, item):
        res_path = bpy.utils.extension_path_user(package = constants.PACKAGE, path = "textures")
        trim_textures = os.path.join(res_path, 'trims')

        layer = "_leggings" if piece == "leggings" else ""

        material = item.material_slots[0].material
        nodes = material.node_tree.nodes
        trim_texture = os.path.join(trim_textures, getattr(self, f'{piece}_trim') + f'{layer}.png')
        image = bpy.data.images.load(trim_texture)
        nodes['trim'].image = image
        nodes['RGB'].outputs[0].default_value = getattr(self,  f'{piece}_trim_colour')
        
    def add_custom_armor(self, part):
        armor_type = bpy.context.window_manager.thomas_rig_legacy.custom_armor[:-4]
        armor_file = os.path.join(constants.ARMOR_PATH, armor_type  + ".blend")

        # append
        directory = os.path.join(armor_file, "Collection")
        filepath = os.path.join(directory, part)
        bpy.ops.wm.append(filepath=filepath, filename=part,directory=directory,
                        link=False, active_collection=True)
        bpy.ops.object.mode_set(mode = "OBJECT")
        items = bpy.context.selected_objects

        for idx, item in enumerate(items):
            # add armature modifier
            utils.add_modifier_armature(self.rig, item)

            # lattice modifier
            if part == "helmet":
                # scuba does not need head lattice for 2nd helmet part but taper
                if armor_type == 'Scuba' and idx == 1:
                    self.add_taper(item, part)
                    continue
                
                utils.add_modifier_lattice_smart_deform(self.rig, item)
                utils.add_modifier_lattice_head(self.rig, item)

            if (armor_type in {'Cultist', 'ZombiePlate'}) and "helmet" in part:
                self.rig.pose.bones["Main_Properties"]["No face"] = True
            elif armor_type in {'Scuba'}:
                # taper lattice
                self.add_taper(item, part)
                self.add_drivers_default_armor(item, self.rig, part)
            elif armor_type == 'Samurai_2' and part == 'leggings' and 'leggings' in item.name:
                self.add_drivers_default_armor(item, self.rig, part)

            # add subdivisions
            utils.add_modifier_subdivision(item, 1, 3)

    def add_taper(self, item, part):
        if "boots" in part or "leggings" in part:
            # right leg
            lattice = self.rig.children[0].modifiers["Lattice_R_Leg"].object
            item.modifiers.new(name = "Taper.R", type = "LATTICE")
            item.modifiers["Taper.R"].object = lattice
            item.modifiers["Taper.R"].vertex_group = "Lattice_Right"

            # left leg
            lattice = self.rig.children[0].modifiers["Lattice_L_Leg"].object
            item.modifiers.new(name = "Taper.L", type = "LATTICE")
            item.modifiers["Taper.L"].object = lattice
            item.modifiers["Taper.L"].vertex_group = "Lattice_Left"

            # move to top
            bpy.context.view_layer.objects.active = item
            bpy.ops.object.modifier_move_to_index(modifier="Taper.R", index=0)
            bpy.ops.object.modifier_move_to_index(modifier="Taper.L", index=1)

            if "leggings" in part:
                lattice = self.rig.children[0].modifiers["Chest"].object
                item.modifiers.new(name = "Chest", type = "LATTICE")
                item.modifiers["Chest"].object = lattice
                item.modifiers["Chest"].vertex_group = "UpperBodyMain"
                # move to top
                bpy.context.view_layer.objects.active = item
                bpy.ops.object.modifier_move_to_index(modifier="Chest", index=2)

        elif "chestplate" in part or "chest_lattice" in item.name:
            # chestplate
            utils.add_modifier_lattice_chest(self.rig, item)
            # move to top
            bpy.context.view_layer.objects.active = item
            bpy.ops.object.modifier_move_to_index(modifier="Chest", index=0)
        return
    
    def add_drivers_default_armor(self, item, rig, part):
        if part.startswith("chestplate"):
            joint = "arm_elbow"
        elif part.startswith("boots") or part.startswith("leggings"):
            joint = "Leg_knee"
        else:
            return
        self.add_driver_bend_120(item, rig, ".R", "Bend.R", joint, "a")
        self.add_driver_bend_120(item, rig, ".R", "120.R", joint, "b")
        self.add_driver_bend_120(item, rig, ".L", "Bend.L", joint, "a")
        self.add_driver_bend_120(item, rig, ".L", "120.L", joint, "b")

    def add_driver_bend_120(self, item, rig, side, shapeKeyName, boneTarget, ab):
        if ab == "a":
            expression = "tan(value*0.5) if tan(value*0.5) >= 0 else -tan(value*0.5)"
        elif ab == "b":
            expression = "tan(value*0.5)-1 if tan(value*0.5) >= 0 else -tan(value*0.5)-1"

        driver = item.data.shape_keys.key_blocks[shapeKeyName].driver_add("value").driver
        driver.type = "SCRIPTED"

        var = driver.variables.new()
        var.type = "TRANSFORMS"
        var.name = "value"

        driver.variables[0].targets[0].id = rig
        driver.variables[0].targets[0].bone_target = boneTarget + side
        driver.variables[0].targets[0].transform_type = "ROT_X"
        driver.variables[0].targets[0].transform_space = "LOCAL_SPACE"
        driver.expression = expression

class ThomasRigBakeArmor:
    def delete_nodes_no_trim(item):
        mat = item.data.materials[0]
        nodes = mat.node_tree.nodes
        keep_nodes = {nodes['Principled BSDF'], nodes['Material Output'], nodes['base']}

        # Remove other nodes
        for node in list(nodes):  # Use list to avoid RuntimeError: Set changed size during iteration
            if node not in keep_nodes:
                nodes.remove(node)
        
        # Connect nodes
        links = mat.node_tree.links
        links.new(nodes['base'].outputs['Color'], nodes['Principled BSDF'].inputs['Base Color'])

        # Move nodes
        nodes['base'].location = (0, 0)
        nodes['Principled BSDF'].location = (400, 0)
        nodes['Material Output'].location = (800, 0)

    def adjust_material(item):
        """adjust material and its node as well as bakes the armor texture, packs img"""
        scn = bpy.context.scene
        mat = item.data.materials[0]
        nodes = mat.node_tree.nodes
        render_engine = scn.render.engine

        # Set up baking settings
        scn.render.engine = 'CYCLES'
        scn.cycles.bake_type = 'DIFFUSE'
        scn.render.bake.use_pass_direct = False
        scn.render.bake.use_pass_indirect = False
        scn.render.bake.use_pass_color = True

        # Create image
        image_name = item.name + '_BakedTexture'
        img = bpy.data.images.new(image_name, 64, 32, alpha=True)

        # Set up baking nodes
        texture_node = nodes.new('ShaderNodeTexImage')
        texture_node.name = 'Bake_Node'
        texture_node.select = True
        texture_node.interpolation = 'Closest'
        nodes.active = texture_node
        texture_node.image = img

        # Bake
        bpy.context.view_layer.objects.active = item
        bpy.ops.object.bake(type='DIFFUSE')
        img.pack()

        # Restore original render engine
        scn.render.engine = render_engine

        # Define nodes to keep
        keep_nodes = {texture_node, nodes['Principled BSDF'], nodes['Material Output'], nodes['trim'], nodes['RGB']}

        # Remove other nodes
        for node in list(nodes):  # Use list to avoid RuntimeError: Set changed size during iteration
            if node not in keep_nodes:
                if node.type == 'TEX_IMAGE' and node.image is not None:
                    bpy.data.images.remove(node.image)
                nodes.remove(node)

        # create new colour nodes
        rgb_node = nodes['RGB']

        mix_1_node = nodes.new('ShaderNodeMixRGB')
        mix_1_node.blend_type = 'MULTIPLY'
        mix_1_node.inputs['Fac'].default_value = 1
        mix_1_node.hide = True

        mix_2_node = nodes.new('ShaderNodeMixRGB')
        mix_2_node.hide = True

        # Move nodes
        nodes['Principled BSDF'].location = (640, 0)
        nodes['Material Output'].location = (920, 0)
        nodes['trim'].location = (0, 40)
        rgb_node.location = (0, -260)
        mix_1_node.location = (280, 40)
        mix_2_node.location = (460, 40) 

        # Connect nodes
        links = mat.node_tree.links
        links.new(texture_node.outputs['Color'], mix_2_node.inputs['Color1'])
        links.new(nodes['trim'].outputs['Color'], mix_1_node.inputs['Color1'])
        links.new(nodes['trim'].outputs['Alpha'], mix_2_node.inputs['Fac'])
        links.new(mix_1_node.outputs['Color'], mix_2_node.inputs['Color2'])
        links.new(rgb_node.outputs['Color'], mix_1_node.inputs['Color2'])

        links.new(mix_2_node.outputs['Color'], nodes['Principled BSDF'].inputs['Base Color'])
    
    def add_nodes_leather(item):
        mat = item.data.materials[0]
        nodes = mat.node_tree.nodes

        armor_trim = 'Bake_Node' in nodes

        # create new nodes
        mix_node = nodes.new('ShaderNodeMixRGB')
        mix_node.blend_type = 'MULTIPLY'
        mix_node.inputs['Color2'].default_value = (0.132891, 0.072269, 0.033102, 1)

        colour_ramp = nodes.new('ShaderNodeValToRGB')
        colour_ramp.width = 140
        colour_ramp.color_ramp.interpolation = 'CONSTANT'
        colour_ramp.color_ramp.elements[1].position = 0.1
        colour_ramp.hide = True
        
        # Connect nodes
        texture_node = nodes['Bake_Node'] if armor_trim else nodes['base']
        links = mat.node_tree.links
        links.new(texture_node.outputs['Color'], colour_ramp.inputs['Fac'])
        links.new(colour_ramp.outputs['Color'], mix_node.inputs['Fac'])
        links.new(texture_node.outputs['Color'], mix_node.inputs['Color1'])
        
        if armor_trim: # if armor trim
            links.new(mix_node.outputs['Color'], nodes['Mix (Legacy).001'].inputs['Color1'])
            mix_node.location = (460, 0)
            colour_ramp.location = (280, 0)
        else:
            links.new(mix_node.outputs['Color'], nodes['Principled BSDF'].inputs['Base Color'])
            texture_node.location = (-240, 0)
            colour_ramp.location = (40, 0)
            mix_node.location = (220, 0)

    def remove_alpha_faces(item):
        # Get the object and its mesh data
        mesh = item.data

        # Create a BMesh from the mesh data
        bm = bmesh.new()
        bm.from_mesh(mesh)

        # Get the UV layer
        uv_layer = bm.loops.layers.uv.verify()

        # Get the image
        nodes = item.data.materials[0].node_tree.nodes
        if 'Bake_Node' in nodes:
            image = nodes['Bake_Node'].image
        else:
            image = nodes['base'].image

        # Calculate the width and height
        width, height = image.size

        # Loop through the faces of the BMesh
        for face in bm.faces:
            # For each face, check if any of its vertices have an alpha value of 0 in the UV map
            for loop in face.loops:
                uv_data = loop[uv_layer].uv
                pixel_index = 4 * (int(uv_data[1]*height) * width + int(uv_data[0]*width))
                pixel = image.pixels[pixel_index:pixel_index+4]
                if pixel[3] == 0:  # If the alpha value is 0
                    # remove the face
                    bmesh.ops.delete(bm, geom=[face], context='FACES')
                    break

        # Update the mesh data with the new BMesh data
        bm.to_mesh(mesh)
        bm.free()